@@ -3,26 +3,27 @@ import asyncio
 import contextlib
 import json
 import logging
-import socket
 import time
 from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
 
-from pywizlight.bulblibrary import BulbClass, BulbType
+from pywizlight.exceptions import WizLightNotKnownBulb
+
+from pywizlight.bulblibrary import BulbType
 from pywizlight.exceptions import (
     WizLightConnectionError,
     WizLightMethodNotFound,
     WizLightTimeOutError,
 )
+from pywizlight.models import DiscoveredBulb
 from pywizlight.protocol import WizProtocol
 from pywizlight.push_manager import PushManager
 from pywizlight.rgbcw import hs2rgbcw, rgb2rgbcw
-from pywizlight.scenes import SCENES
-from pywizlight.utils import hex_to_percent, percent_to_hex
+from pywizlight.scenes import SCENES, SCENES_BY_CLASS
+from pywizlight.utils import hex_to_percent, percent_to_hex, to_wiz_json
 from pywizlight.vec import Vector
 
 _LOGGER = logging.getLogger(__name__)
-TW_SCENES = [6, 9, 10, 11, 12, 13, 14, 15, 16, 18, 29, 30, 31, 32]
-DW_SCENES = [9, 10, 13, 14, 29, 30, 31, 32]
+
 
 BulbResponse = Dict[str, Any]
 
@@ -37,17 +38,56 @@ MAX_TIME_BETWEEN_PUSH = PUSH_KEEP_ALIVE_INTERVAL + TIMEOUT
 
 NEVER_TIME = -120.0
 
-_IGNORE_KEYS = {"src", "mqttCd", "ts", "rssi"}
+
+RGB_ORDER = ["r", "g", "b"]
+RGB_COLORS = set(RGB_ORDER)
+RGBW_ORDER = ["r", "g", "b", "w"]
+RGBW_COLORS = set(RGBW_ORDER)
+RGBWW_ORDER = ["r", "g", "b", "c", "w"]
+RGBWW_COLORS = set(RGBWW_ORDER)
+
+_IGNORE_STATE_KEYS = {"mqttCd", "ts", "rssi"}
+PIR_SOURCE = "pir"
+
+WIZMOTE_BUTTON_MAP = {
+    "wfa1": "on",
+    "wfa2": "off",
+    "wfa3": "night",
+    "wfa8": "decrease_brightness",
+    "wfa9": "increase_brightness",
+    "wfa16": "1",
+    "wfa17": "2",
+    "wfa18": "3",
+    "wfa19": "4",
+}
+
+ALWAYS_SEND_SRCS = set([PIR_SOURCE, *WIZMOTE_BUTTON_MAP])
 
 
-def states_match(old: dict[str, Any], new: dict[str, Any]) -> bool:
+def states_match(old: Dict[str, Any], new: Dict[str, Any]) -> bool:
     """Check if states match except for keys we do not want to callback on."""
+    old_src = old.get("src")
+    new_src = new.get("src")
+    # Always send the update when there is a PIR change or button press.
+    if new_src != old_src and (
+        new_src in ALWAYS_SEND_SRCS or old_src in ALWAYS_SEND_SRCS
+    ):
+        return False
     for key, val in new.items():
-        if old.get(key) != val and key not in _IGNORE_KEYS:
+        if key != "src" and old.get(key) != val and key not in _IGNORE_STATE_KEYS:
             return False
     return True
 
 
+def _rgb_in_range_or_raise(rgb: Tuple[Union[float, int], ...]) -> None:
+    if not (0 <= rgb[0] < 256):
+        raise ValueError("Red is not in range between 0-255.")
+    if not (0 <= rgb[1] < 256):
+        raise ValueError("Green is not in range between 0-255.")
+    if not (0 <= rgb[2] < 256):
+        raise ValueError("Blue is not in range between 0-255.")
+
+
 class PilotBuilder:
     """Get information from the bulb."""
 
@@ -58,6 +98,8 @@ class PilotBuilder:
         speed: Optional[int] = None,
         scene: Optional[int] = None,
         rgb: Optional[Tuple[float, float, float]] = None,
+        rgbw: Optional[Tuple[int, int, int, int]] = None,
+        rgbww: Optional[Tuple[int, int, int, int, int]] = None,
         hucolor: Optional[Tuple[float, float]] = None,
         brightness: Optional[int] = None,
         colortemp: Optional[int] = None,
@@ -65,82 +107,86 @@ class PilotBuilder:
     ) -> None:
         """Set the parameter."""
         self.pilot_params: Dict[str, Any] = {"state": state}
-
-        if warm_white is not None:
-            self._set_warm_white(warm_white)
-        if cold_white is not None:
-            self._set_cold_white(cold_white)
         if speed is not None:
             self._set_speed(speed)
         if scene is not None:
             self._set_scene(scene)
-        if rgb is not None:
-            self._set_rgb(rgb)
         if brightness is not None:
             self._set_brightness(brightness)
-        if colortemp is not None:
+        if rgb is not None:
+            self._set_rgb(rgb)
+        elif rgbw is not None:
+            self._set_rgbw(rgbw)
+        elif rgbww is not None:
+            self._set_rgbww(rgbww)
+        elif colortemp is not None:
             self._set_colortemp(colortemp)
-        if hucolor is not None:
+        elif hucolor is not None:
             self._set_hs_color(hucolor)
+        if warm_white is not None:
+            self._set_warm_white(warm_white)
+        if cold_white is not None:
+            self._set_cold_white(cold_white)
 
     def set_pilot_message(self) -> str:
         """Return the pilot message."""
-        return json.dumps({"method": "setPilot", "params": self.pilot_params})
+        return to_wiz_json({"method": "setPilot", "params": self.pilot_params})
 
     def set_state_message(self, state: bool) -> str:
         """Return the setState message. It doesn't change the current status of the light."""
         self.pilot_params["state"] = state
-        return json.dumps({"method": "setState", "params": self.pilot_params})
+        return to_wiz_json({"method": "setState", "params": self.pilot_params})
 
     def _set_warm_white(self, value: int) -> None:
         """Set the value of the warm white led."""
-        if 0 <= value < 256:
-            self.pilot_params["w"] = value
-        else:
+        if not 0 <= value < 256:
             raise ValueError("Value must be between 0 and 255")
+        self.pilot_params["w"] = value
 
     def _set_cold_white(self, value: int) -> None:
         """Set the value of the cold white led."""
-        if 0 <= value < 256:
-            self.pilot_params["c"] = value
-        else:
+        if not 0 <= value < 256:
             raise ValueError("Value must be between 0 and 255")
+        self.pilot_params["c"] = value
 
     def _set_speed(self, value: int) -> None:
         """Set the color changing speed in precent (0-100)."""
         # This applies only to changing effects.
-        if 0 < value < 101:
-            self.pilot_params["speed"] = value
-        else:
+        if not 0 < value < 101:
             raise ValueError("Value must be between 0 and 100")
+        self.pilot_params["speed"] = value
 
     def _set_scene(self, scene_id: int) -> None:
         """Set the scene by id."""
-        if scene_id in SCENES:
-            self.pilot_params["sceneId"] = scene_id
-        else:
+        if scene_id not in SCENES:
             # id not in SCENES !
             raise ValueError("Scene is not available. Only 0 to 32 are supported")
+        self.pilot_params["sceneId"] = scene_id
+
+    def _set_rgbw(self, rgbw: Tuple[int, int, int, int]) -> None:
+        """Set the RGBW color state of the bulb."""
+        _rgb_in_range_or_raise(rgbw)
+        self.pilot_params.update({key: rgbw[idx] for idx, key in enumerate(RGB_ORDER)})
+        self._set_warm_white(rgbw[3])
+
+    def _set_rgbww(self, rgbww: Tuple[int, int, int, int, int]) -> None:
+        """Set the RGBWW color state of the bulb."""
+        _rgb_in_range_or_raise(rgbww)
+        params = self.pilot_params
+        params.update({key: rgbww[idx] for idx, key in enumerate(RGB_ORDER)})
+        self._set_cold_white(rgbww[3])
+        self._set_warm_white(rgbww[4])
 
     def _set_rgb(self, values: Tuple[float, float, float]) -> None:
         """Set the RGB color state of the bulb."""
-
         # Setup the tuples for the RGB values
         red, green, blue = values
-        if not (0 <= red < 256):
-            raise ValueError("Red is not in range between 0-255.")
-        if not (0 <= green < 256):
-            raise ValueError("Green is not in range between 0-255.")
-        if not (0 <= blue < 256):
-            raise ValueError("Blue is not in range between 0-255.")
+        _rgb_in_range_or_raise(values)
         # Get the RGB+CW values
         rgb_out, cw = rgb2rgbcw(values)
         # Extract the RGB values
-        red, green, blue = rgb_out
-        # Set the RGB values
-        self.pilot_params["r"] = red
-        self.pilot_params["g"] = green
-        self.pilot_params["b"] = blue
+        params = self.pilot_params
+        params.update({key: rgb_out[idx] for idx, key in enumerate(RGB_ORDER)})
         # No CW because of full RGB color
         if cw is not None:
             # Use the existing set_warm_white function to set the CW values
@@ -150,17 +196,10 @@ class PilotBuilder:
         """Set the HS color state of the bulb."""
         # Transform the HS values to RGB+CW values
         rgb, cw = hs2rgbcw(values)
-        red, green, blue = rgb
-        if not (0 <= red < 256):
-            raise ValueError("Red is not in range between 0-255.")
-        if not (0 <= green < 256):
-            raise ValueError("Green is not in range between 0-255.")
-        if not (0 <= blue < 256):
-            raise ValueError("Blue is not in range between 0-255.")
+        _rgb_in_range_or_raise(rgb)
         # Set the RGB values
-        self.pilot_params["r"] = red
-        self.pilot_params["g"] = green
-        self.pilot_params["b"] = blue
+        params = self.pilot_params
+        params.update({key: rgb[idx] for idx, key in enumerate(RGB_ORDER)})
         if cw is not None:
             # Use the existing set_warm_white function to set the CW values
             self._set_warm_white(cw)
@@ -168,22 +207,27 @@ class PilotBuilder:
     def _set_brightness(self, value: int) -> None:
         """Set the value of the brightness 0-255."""
         percent = hex_to_percent(value)
-        # hardware limitation - values less than 10% are not supported
-        if percent < 10:
-            percent = 10
         if percent > 101:
             raise ValueError("Max value can be 100% with 255.")
-        self.pilot_params["dimming"] = percent
+        # hardware limitation - values less than 10% are not supported
+        self.pilot_params["dimming"] = max(10, percent)
 
     def _set_colortemp(self, kelvin: int) -> None:
         """Set the color temperature for the white led in the bulb."""
         # normalize the kelvin values - should be removed
-        if kelvin < 1000:
-            kelvin = 1000
-        if kelvin > 10000:
-            kelvin = 10000
+        self.pilot_params["temp"] = min(10000, max(1000, kelvin))
 
-        self.pilot_params["temp"] = kelvin
+
+def _extract_bool(response: BulbResponse, key: str) -> Optional[bool]:
+    return bool(response[key]) if key in response else None
+
+
+def _extract_str(response: BulbResponse, key: str) -> Optional[str]:
+    return str(response[key]) if key in response else None
+
+
+def _extract_int(response: BulbResponse, key: str) -> Optional[int]:
+    return int(response[key]) if key in response else None
 
 
 class PilotParser:
@@ -195,31 +239,25 @@ class PilotParser:
 
     def get_state(self) -> Optional[bool]:
         """Return the state of the bulb."""
-        if "state" in self.pilotResult:
-            return bool(self.pilotResult["state"])
-        else:
-            return None
+        return _extract_bool(self.pilotResult, "state")
+
+    def get_source(self) -> Optional[str]:
+        """Return the source of the state change."""
+        return _extract_str(self.pilotResult, "src")
 
     def get_mac(self) -> Optional[str]:
         """Return MAC from the bulb."""
-        if "mac" in self.pilotResult:
-            return str(self.pilotResult["mac"])
-        else:
-            return None
+        return _extract_str(self.pilotResult, "mac")
 
     def get_warm_white(self) -> Optional[int]:
         """Get the value of the warm white led."""
-        if "w" in self.pilotResult:
-            return int(self.pilotResult["w"])
-        else:
-            return None
+        return _extract_int(self.pilotResult, "w")
 
     def get_white_range(self) -> Optional[List[float]]:
         """Get the value of the whiteRange property."""
         if "whiteRange" in self.pilotResult:
             return [float(x) for x in self.pilotResult["whiteRange"]]
-        else:
-            return None
+        return None
 
     def get_extended_white_range(self) -> Optional[List[float]]:
         """Get the value of the extended whiteRange property."""
@@ -228,48 +266,50 @@ class PilotParser:
         # New after v1.22 FW - "cctRange":[2200,2700,6500,6500]
         elif "cctRange" in self.pilotResult:
             return [float(x) for x in self.pilotResult["cctRange"]]
-        else:
-            return None
+        return None
 
     def get_speed(self) -> Optional[int]:
         """Get the color changing speed."""
-        if "speed" in self.pilotResult:
-            return int(self.pilotResult["speed"])
-        else:
-            return None
+        return _extract_int(self.pilotResult, "speed")
 
     def get_scene(self) -> Optional[str]:
         """Get the current scene name."""
         if "schdPsetId" in self.pilotResult:  # rhythm
             return SCENES[1000]
-
-        scene_id = self.pilotResult["sceneId"]
-        if scene_id in SCENES:
-            return SCENES[scene_id]
-        else:
+        if "sceneId" not in self.pilotResult:
             return None
+        return SCENES.get(self.pilotResult["sceneId"])
 
     def get_cold_white(self) -> Optional[int]:
         """Get the value of the cold white led."""
-        if "c" in self.pilotResult:
-            return int(self.pilotResult["c"])
-        else:
-            return None
+        return _extract_int(self.pilotResult, "c")
 
     def get_rgb(self) -> Union[Tuple[None, None, None], Vector]:
         """Get the RGB color state of the bulb and turns it on."""
-        if (
-            "r" in self.pilotResult
-            and "g" in self.pilotResult
-            and "b" in self.pilotResult
-        ):
-            r = self.pilotResult["r"]
-            g = self.pilotResult["g"]
-            b = self.pilotResult["b"]
-            return float(r), float(g), float(b)
-        else:
-            # no RGB color value was set
-            return None, None, None
+        state = self.pilotResult
+        if RGB_COLORS.issubset(state):
+            return tuple(int(state[val]) for val in RGB_ORDER)
+        # no RGB color value was set
+        return None, None, None
+
+    def get_rgbw(self) -> Optional[Tuple[int, int, int, int]]:
+        """Get the RGBW color of the device."""
+        state = self.pilotResult
+        if not RGBW_COLORS.issubset(state):
+            return None
+        return cast(
+            Tuple[int, int, int, int], tuple(int(state[val]) for val in RGBW_ORDER)
+        )
+
+    def get_rgbww(self) -> Optional[Tuple[int, int, int, int, int]]:
+        """Get the RGBWW color of the device."""
+        state = self.pilotResult
+        if not RGBWW_COLORS.issubset(state):
+            return None
+        return cast(
+            Tuple[int, int, int, int, int],
+            tuple(int(state[val]) for val in RGBWW_ORDER),
+        )
 
     def get_brightness(self) -> Optional[int]:
         """Get the value of the brightness 0-255."""
@@ -279,10 +319,25 @@ class PilotParser:
 
     def get_colortemp(self) -> Optional[int]:
         """Get the color temperature from the bulb."""
-        if "temp" in self.pilotResult:
-            return int(self.pilotResult["temp"])
-        else:
-            return None
+        return _extract_int(self.pilotResult, "temp")
+
+
+async def _send_udp_message_with_retry(
+    message: str,
+    transport: asyncio.DatagramTransport,
+    response_future: asyncio.Future,
+    ip: str,
+    port: int,
+) -> None:
+    """Send a UDP message multiple times until we reach the maximum or a response is recieved."""
+    data = message.encode("utf-8")
+    for send in range(MAX_SEND_DATAGRAMS):
+        if transport.is_closing() or response_future.done():
+            return
+        attempt = send + 1
+        _LOGGER.debug("%s: >> %s (%s/%s)", ip, data, attempt, MAX_SEND_DATAGRAMS)
+        transport.sendto(data, (ip, port))
+        await asyncio.sleep(SEND_INTERVAL)
 
 
 class wizlight:
@@ -293,7 +348,6 @@ class wizlight:
     def __init__(
         self,
         ip: str,
-        connect_on_init: bool = False,
         port: int = 38899,
         mac: Optional[str] = None,
     ) -> None:
@@ -303,6 +357,7 @@ class wizlight:
         self.state: Optional[PilotParser] = None
         self.mac = mac
         self.bulbtype: Optional[BulbType] = None
+        self.modelConfig: Optional[Dict] = None
         self.whiteRange: Optional[List[float]] = None
         self.extwhiteRange: Optional[List[float]] = None
         self.transport: Optional[asyncio.DatagramTransport] = None
@@ -310,14 +365,12 @@ class wizlight:
 
         self.lock = asyncio.Lock()
         self.loop = asyncio.get_event_loop()
-        self.push_callback: Optional[Callable] = None
+        self.push_callback: Optional[Callable[[PilotParser], None]] = None
         self.response_future: Optional[asyncio.Future] = None
         self.push_cancel: Optional[Callable] = None
         self.last_push: float = NEVER_TIME
         self.push_running: bool = False
-        # check the state on init
-        if connect_on_init:
-            self._check_connection()
+        # Check connection removed as it did blocking I/O in the event loop
 
     @property
     def status(self) -> Optional[bool]:
@@ -333,7 +386,7 @@ class wizlight:
         if self.transport:
             return
         transport_proto = await self.loop.create_datagram_endpoint(
-            lambda: WizProtocol(on_response=self._on_response),
+            lambda: WizProtocol(on_response=self._on_response, on_error=self._on_error),
             remote_addr=(self.ip, self.port),
         )
         self.transport = cast(asyncio.DatagramTransport, transport_proto[0])
@@ -342,17 +395,21 @@ class wizlight:
     def register(self) -> None:
         """Call register to keep alive push updates."""
         if self.push_running:
-            asyncio.ensure_future(
+            asyncio.create_task(
                 self._async_send_register(PushManager().get().register_msg)
             )
             self.loop.call_later(PUSH_KEEP_ALIVE_INTERVAL, self.register)
 
     async def _async_send_register(self, message: str) -> None:
         """Send the registration message."""
-        with contextlib.suppress(WizLightTimeOutError):
+        try:
             await self.sendUDPMessage(message)
+        except (WizLightTimeOutError, WizLightConnectionError) as ex:
+            _LOGGER.debug("%s: Registration for push updates failed: %s", self.ip, ex)
 
-    async def start_push(self, callback: Callable) -> None:
+    async def start_push(
+        self, callback: Optional[Callable[[PilotParser], None]]
+    ) -> None:
         """Start periodic register calls to get push updates via syncPilot."""
         _LOGGER.debug("Enabling push updates for %s", self.mac)
         self.push_callback = callback
@@ -362,6 +419,12 @@ class wizlight:
             self.push_running = True
             self.register()
 
+    def set_discovery_callback(
+        self, callback: Optional[Callable[[DiscoveredBulb], None]]
+    ) -> None:
+        """Set the callback for when a new device is discovered."""
+        PushManager().get().set_discovery_callback(callback)
+
     def _on_push(self, resp: dict, addr: Tuple[str, int]) -> None:
         """Handle a syncPilot from the device."""
         self.last_push = time.monotonic()
@@ -379,22 +442,10 @@ class wizlight:
         if self.response_future and not self.response_future.done():
             self.response_future.set_result(message)
 
-    def _check_connection(self) -> None:
-        """Check the connection to the bulb."""
-        message = r'{"method":"getPilot","params":{}}'
-        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-        sock.settimeout(2)
-        try:
-            # send a udp package
-            sock.sendto(bytes(message, "utf-8"), (self.ip, self.port))
-            # get response data
-            data, addr = sock.recvfrom(1024)
-            if data:
-                return
-        except socket.timeout:
-            raise WizLightTimeOutError(
-                "No connection was established by initialization."
-            )
+    def _on_error(self, exception: Optional[Exception]) -> None:
+        """Handle a protocol error."""
+        if exception and self.response_future and not self.response_future.done():
+            self.response_future.set_exception(exception)
 
     async def get_bulbtype(self) -> BulbType:
         """Return the bulb type as BulbType object."""
@@ -404,11 +455,23 @@ class wizlight:
         bulb_config = await self.getBulbConfig()
         result = bulb_config["result"]
         if "moduleName" not in result:
-            raise ValueError("Unable to determine bulb type.")
+            raise WizLightNotKnownBulb(
+                "Unable to determine the bulb type; moduleName is missing from getSystemConfig"
+            )
         white_range = await self.getExtendedWhiteRange()
+        white_to_color_ratio = None
+        white_channels = None
+        if "drvConf" in result:
+            # For old FW < 1.22
+            white_to_color_ratio, white_channels = result["drvConf"]
         module_name = result["moduleName"]
         fw_version = result.get("fwVersion")
-        self.bulbtype = BulbType.from_data(module_name, white_range, fw_version)
+        model_result = self.modelConfig["result"] if self.modelConfig else {}
+        white_channels = model_result.get("nowc", white_channels)
+        white_to_color_ratio = model_result.get("wcr", white_to_color_ratio)
+        self.bulbtype = BulbType.from_data(
+            module_name, white_range, fw_version, white_channels, white_to_color_ratio
+        )
         return self.bulbtype
 
     async def getWhiteRange(self) -> Optional[List[float]]:
@@ -429,15 +492,13 @@ class wizlight:
 
         # First for FW > 1.22
         resp = await self.getModelConfig()
-        if resp is not None and "result" in resp:
-            self.extwhiteRange = PilotParser(resp["result"]).get_extended_white_range()
-        else:
+        if resp is None or "result" not in resp:
             # For old FW < 1.22
             resp = await self.getUserConfig()
-            if "result" in resp:
-                self.extwhiteRange = PilotParser(
-                    resp["result"]
-                ).get_extended_white_range()
+
+        if "result" in resp:
+            self.extwhiteRange = PilotParser(resp["result"]).get_extended_white_range()
+
         return self.extwhiteRange
 
     async def getSupportedScenes(self) -> List[str]:
@@ -448,29 +509,19 @@ class wizlight:
         if self.bulbtype is None:
             await self.get_bulbtype()
         assert self.bulbtype  # Should have gotten set by get_bulbtype
-        # return for TW
-        if self.bulbtype.bulb_type == BulbClass.TW:
-            return [SCENES[key] for key in TW_SCENES]
-        # retrun for DW
-        if self.bulbtype.bulb_type == BulbClass.DW:
-            return [SCENES[key] for key in DW_SCENES]
-        # return for RGB - all scenes supported
-        return list(SCENES.values())
+        return SCENES_BY_CLASS.get(self.bulbtype.bulb_type, [])
 
     async def turn_off(self) -> None:
         """Turn the light off."""
-        message = r'{"method":"setPilot","params":{"state":false}}'
-        await self.sendUDPMessage(message)
+        await self.sendUDPMessage(r'{"method":"setPilot","params":{"state":false}}')
 
     async def reboot(self) -> None:
         """Reboot the bulb."""
-        message = r'{"method":"reboot","params":{}}'
-        await self.sendUDPMessage(message)
+        await self.sendUDPMessage(r'{"method":"reboot","params":{}}')
 
     async def reset(self) -> None:
         """Reset the bulb to factory defaults."""
-        message = r'{"method":"reset","params":{}}'
-        await self.sendUDPMessage(message)
+        await self.sendUDPMessage(r'{"method":"reset","params":{}}')
 
     async def turn_on(self, pilot_builder: PilotBuilder = PilotBuilder()) -> None:
         """Turn the light on with defined message.
@@ -499,8 +550,7 @@ class wizlight:
         {"method": "getPilot", "id": 24}
         """
         if self.last_push + MAX_TIME_BETWEEN_PUSH < time.monotonic():
-            message = r'{"method":"getPilot","params":{}}'
-            resp = await self.sendUDPMessage(message)
+            resp = await self.sendUDPMessage(r'{"method":"getPilot","params":{}}')
             if resp is not None and "result" in resp:
                 self.state = PilotParser(resp["result"])
             else:
@@ -520,8 +570,7 @@ class wizlight:
 
     async def getBulbConfig(self) -> BulbResponse:
         """Return the configuration from the bulb."""
-        message = r'{"method":"getSystemConfig","params":{}}'
-        resp = await self.sendUDPMessage(message)
+        resp = await self.sendUDPMessage(r'{"method":"getSystemConfig","params":{}}')
         self._cache_mac_from_bulb_config(resp)
         return resp
 
@@ -529,16 +578,16 @@ class wizlight:
         """Return the new model capabilities from the bulb.
         Only available in bulb FW >1.22!
         """
-        try:
-            message = r'{"method":"getModelConfig","params":{}}'
-            return await self.sendUDPMessage(message)
-        except WizLightMethodNotFound:
-            return None
+        if self.modelConfig is None:
+            with contextlib.suppress(WizLightMethodNotFound):
+                self.modelConfig = await self.sendUDPMessage(
+                    r'{"method":"getModelConfig","params":{}}'
+                )
+        return self.modelConfig
 
     async def getUserConfig(self) -> BulbResponse:
         """Return the user configuration from the bulb."""
-        message = r'{"method":"getUserConfig","params":{}}'
-        return await self.sendUDPMessage(message)
+        return await self.sendUDPMessage(r'{"method":"getUserConfig","params":{}}')
 
     async def lightSwitch(self) -> None:
         """Turn the light bulb on or off like a switch."""
@@ -556,40 +605,40 @@ class wizlight:
     async def sendUDPMessage(self, message: str) -> BulbResponse:
         """Send the UDP message to the bulb."""
         await self._ensure_connection()
-        data = message.encode("utf-8")
         assert self.transport is not None
         async with self.lock:
             self.response_future = asyncio.Future()
-            for send in range(MAX_SEND_DATAGRAMS):
-                attempt = send + 1
-                _LOGGER.debug(
-                    "%s: >> %s (%s/%s)", self.ip, data, attempt, MAX_SEND_DATAGRAMS
+            send_task = asyncio.create_task(
+                _send_udp_message_with_retry(
+                    message, self.transport, self.response_future, self.ip, self.port
                 )
-                self.transport.sendto(data, (self.ip, self.port))
-                try:
-                    response = await asyncio.wait_for(
-                        asyncio.shield(self.response_future), timeout=SEND_INTERVAL
-                    )
-                except asyncio.TimeoutError:
-                    _LOGGER.debug(
-                        "%s: Timed out waiting for response to %s (%s/%s)",
-                        self.ip,
-                        message,
-                        attempt,
-                        MAX_SEND_DATAGRAMS,
-                    )
-                    if attempt == MAX_SEND_DATAGRAMS:
-                        raise WizLightTimeOutError("The request to the bulb timed out")
-                else:
-                    break
-
+            )
+            try:
+                response = await asyncio.wait_for(
+                    asyncio.shield(self.response_future), timeout=TIMEOUT
+                )
+            except OSError as ex:
+                raise WizLightConnectionError(str(ex)) from ex
+            except asyncio.TimeoutError:
+                _LOGGER.debug(
+                    "%s: Timed out waiting for response to %s after %s tries.",
+                    self.ip,
+                    message,
+                    MAX_SEND_DATAGRAMS,
+                )
+                raise WizLightTimeOutError("The request to the bulb timed out")
+            finally:
+                await asyncio.sleep(0)  # Try to let the task finish without cancelation
+                if not send_task.done():
+                    send_task.cancel()
+                    with contextlib.suppress(asyncio.CancelledError):
+                        await send_task
         resp = json.loads(response.decode())
-        if "error" not in resp:
-            return resp
-        elif resp["error"]["code"] == -32601:
-            raise WizLightMethodNotFound("Method not found; maybe older bulb FW?")
-        else:
+        if "error" in resp:
+            if resp["error"]["code"] == -32601:
+                raise WizLightMethodNotFound("Method not found; maybe older bulb FW?")
             raise WizLightConnectionError(f'Error recieved: {resp["error"]}')
+        return resp
 
     async def async_close(self):
         """Close the transport."""
@@ -606,5 +655,5 @@ class wizlight:
 
     def __del__(self):
         """Close the connection when the object is destroyed."""
-        if self.transport:
+        if self.transport and not self.transport.is_closing():
             self.loop.call_soon_threadsafe(self._async_close)
